@@ -1,19 +1,37 @@
 module GPGME
   class Ctx
     IMPORT_FLAGS = {
-      'new key' => 1,
+      'new_key' => 1,
       'new_uids' => 2,
       'new_signatures' => 4,
       'new_subkeys' => 8
     }
 
-    def keyimport(*args)
-      self.import_keys(*args)
+    def keyimport(keydata)
+      self.import_keys(GPGME::Data.new(keydata))
       result = self.import_result
       result.imports.map(&:set_action)
       result
     end
 
+    def find_keys(input=nil, secret_only=nil)
+      _, input = clean_and_classify_input(input)
+      keys(input, secret_only)
+    end
+
+    def clean_and_classify_input(input)
+      case input
+      when /.*?([^ <>]+@[^ <>]+).*?/
+        [:email, "<#{$1}>"]
+      when /^http/
+        [:url, input]
+      when Conf::FINGERPRINT_REGEXP
+        [:fingerprint, "0x#{input.gsub(/^0x/, '')}"]
+      else
+        [nil, input]
+      end
+    end
+
     # Tell gpgme to use the given binary.
     def self.set_gpg_path_from_env
       path = ENV['GPGBIN'].to_s
@@ -42,42 +60,93 @@ module GPGME
       GPGME::Engine.info.find {|e| e.protocol == GPGME::PROTOCOL_OpenPGP }
     end
 
-    def self.refresh_keys(keys)
-      output = []
-      base_args = "--quiet --no-auto-check-trustdb --keyserver #{Conf.keyserver} --refresh-keys"
-      keys.each do |key|
-        args = "#{base_args} #{key.fingerprint}"
-        err, gpgout, _ = gpgcli(args)
-        gpgout = filter_gpgcli_output(gpgout)
-        output << filter_gpgcli_output(err)
-        # Add any gpgkeys-message (gpg 2.0 writes those messages to stdout).
-        # Those could e.g. report a failure to connect to the keyserver.
-        output << gpgout.select { |line| line.match(/^gpgkeys: .*$/) }
-
-        import_stats = translate_import_data(gpgout)
-        if import_stats.present?
-          output << I18n.t("key_updated", { fingerprint: key.fingerprint,
-                                            states: import_stats.join(', ') })
-        end
+    def refresh_keys(keys)
+      output = keys.map do |key|
+        # Sleep a short while to make traffic analysis less easy.
         sleep rand(1.0..5.0)
+        refresh_key(key.fingerprint).presence
       end
-      GPGME::Ctx.gpgcli("--check-trustdb")
-      output.flatten.uniq.join
+      output.compact.join("\n")
     end
 
-    def self.translate_import_data(gpgoutput)
-      result = []
-      import_ok = gpgoutput.grep(/IMPORT_OK/).first
-      return result if import_ok.blank?
+    def refresh_key(fingerprint)
+      args = "--keyserver #{Conf.keyserver} --refresh-keys #{fingerprint}"
+      gpgerr, gpgout, exitcode = self.class.gpgcli(args)
 
-      import_status = import_ok.split(/\s/).slice(2).to_i
-      return result if import_status.zero?
-
-      # TODO: Raise alarm if new key is found?
-      IMPORT_FLAGS.each do |text, int|
-        if (import_status & int) > 0
-          result << I18n.t("import_states.#{text}")
+      if exitcode > 0
+        # Return filtered error messages. Include gpgkeys-messages from stdout
+        # (gpg 2.0 does that), which could e.g. report a failure to connect to
+        # the keyserver.
+        [
+          refresh_key_filter_messages(gpgerr),
+          refresh_key_filter_messages(gpgout).grep(/^gpgkeys: /)
+        ].flatten.compact.join("\n")
+      else
+        lines = translate_output('key_updated', gpgout).reject do |line|
+          # Reduce the noise a little.
+          line.match(/.* \(unchanged\)\.$/)
         end
+        lines.join("\n")
+      end
+    end
+
+    def fetch_key(input)
+      arguments, error = fetch_key_gpg_arguments_for(input)
+      return error if error
+
+      gpgerr, gpgout, exitcode = self.class.gpgcli(arguments)
+
+      # Unfortunately gpg doesn't exit with code > 0 if `--fetch-key` fails.
+      if exitcode > 0 || gpgerr.grep(/ unable to fetch /).presence
+        "Fetching #{input} did not succeed:\n#{gpgerr.join("\n")}"
+      else
+        translate_output('key_fetched', gpgout).join("\n")
+      end
+    end
+
+    def fetch_key_gpg_arguments_for(input)
+      case input
+      when Conf::FINGERPRINT_REGEXP
+        "--keyserver #{Conf.keyserver} --recv-key #{input}"
+      when /^http/
+        "--fetch-key #{input}"
+      when /@/
+        # --recv-key doesn't work with email-addresses, so we use --locate-key
+        # restricted to keyservers.
+        "--keyserver #{Conf.keyserver} --auto-key-locate keyserver --locate-key #{input}"
+      else
+        [nil, I18n.t("fetch_key.invalid_input")]
+      end
+    end
+
+    def translate_output(locale_key, gpgoutput)
+      import_states = translate_import_data(gpgoutput)
+      strings = import_states.map do |fingerprint, states|
+        I18n.t(locale_key, { fingerprint: fingerprint,
+                             states: states.join(', ') })
+      end
+      strings
+    end
+
+    def translate_import_data(gpgoutput)
+      result = {}
+      gpgoutput.grep(/IMPORT_OK/) do |import_ok|
+        next if import_ok.blank?
+
+        import_status, fingerprint = import_ok.split(/\s/).slice(2, 2)
+        import_status = import_status.to_i
+        states = []
+
+        if import_status == 0
+          states << I18n.t("import_states.unchanged")
+        else
+          IMPORT_FLAGS.each do |text, int|
+            if (import_status & int) > 0
+              states << I18n.t("import_states.#{text}")
+            end
+          end
+        end
+        result[fingerprint] = states
       end
       result
     end
@@ -85,7 +154,7 @@ module GPGME
     # Unfortunately we can't distinguish between a failure to connect the
     # keyserver, and a failure to find the key on the server. So we try to
     # filter misleading errors to check if there are any to be reported.
-    def self.filter_gpgcli_output(strings)
+    def refresh_key_filter_messages(strings)
       strings.reject do |line|
         line.chomp == 'gpg: keyserver refresh failed: No data' ||
           line.match(/^gpgkeys: key .* not found on keyserver/) ||
@@ -108,7 +177,7 @@ module GPGME
         else
           output = stdout.readlines
         end
-        stdin.close
+        stdin.close if ! stdin.closed?
         errors = stderr.readlines
         exitcode = thread.value.exitstatus
       end
