@@ -66,6 +66,8 @@ module Schleuder
                 with: /\A[[:graph:]\s]*\z/i,
               }
 
+    default_scope { order(:email) }
+
     def self.configurable_attributes
       @configurable_attributes ||= begin
         all = self.validators.map(&:attributes).flatten.uniq.compact.sort
@@ -89,28 +91,37 @@ module Schleuder
       subscriptions.where(admin: true)
     end
 
+    def subscriptions_without_fingerprint
+      subscriptions.without_fingerprint
+    end
+
     def key(fingerprint=self.fingerprint)
       keys(fingerprint).first
     end
 
     def secret_key
-      gpg.keys(self.fingerprint, true).first
+      keys(self.fingerprint, true).first
     end
 
-    def keys(identifier='')
-      gpg.keys(identifier)
+    def keys(identifier=nil, secret_only=nil)
+      gpg.find_keys(identifier, secret_only)
     end
 
-    def keys_by_email(address)
-      keys("<#{address}>")
+    def distinct_key(identifier)
+      keys = keys(identifier).select { |key| key.usable_for?(:encrypt) }
+      if keys.size == 1
+        return keys.first
+      else
+        return nil
+      end
     end
 
     def import_key(importable)
-      gpg.keyimport(GPGME::Data.new(importable))
+      gpg.keyimport(importable)
     end
 
     def delete_key(fingerprint)
-      if key = gpg.keys(fingerprint).first
+      if key = keys(fingerprint).first
         key.delete!
         true
       else
@@ -152,6 +163,7 @@ module Schleuder
                           fingerprint: key.fingerprint,
                           email: key.email
                       })
+        text << "\n"
       end
 
       unusable.each do |key,usability_issue|
@@ -160,12 +172,30 @@ module Schleuder
                           fingerprint: key.fingerprint,
                           email: key.email
                       })
+        text << "\n"
       end
       text
     end
 
     def refresh_keys
-      GPGME::Ctx.refresh_keys(self.keys)
+      gpg.refresh_keys(self.keys)
+    end
+
+    def fetch_keys(input)
+      gpg.fetch_key(input)
+    end
+
+    def pin_keys
+      updated_emails = subscriptions_without_fingerprint.collect do |subscription|
+        key = distinct_key(subscription.email)
+        if key
+          subscription.update(fingerprint: key.fingerprint)
+          "#{subscription.email}: #{key.fingerprint}"
+        else
+          nil
+        end
+      end
+      updated_emails.compact.join("\n")
     end
 
     def self.by_recipient(recipient)
@@ -231,10 +261,11 @@ module Schleuder
       @listdir ||= self.class.listdir(self.email)
     end
 
+    # TODO: get rid of this method in the future
     def subscribe(email, fingerprint=nil, adminflag=false, deliveryflag=true)
       # Ensure we have true or false as values for these two attributes.
-      admin            = adminflag.to_s == 'true'
-      delivery_enabled = deliveryflag.to_s != 'false'
+      admin            =  (['true', '1'].include?  adminflag.to_s)
+      delivery_enabled = !(['false', '0'].include? deliveryflag.to_s)
 
       sub = Subscription.new(
           list_id: self.id,
@@ -285,6 +316,29 @@ module Schleuder
       self.send("#{attrib}=", value)
     end
 
+    def send_list_key_to_subscriptions
+      mail = Mail.new
+      mail.from = self.email
+      mail.subject = I18n.t('list_public_key_subject')
+      mail.body = I18n.t('list_public_key_attached')
+      mail.attach_list_key!(self)
+      send_to_subscriptions(mail)
+      true
+    end
+
+    def send_to_subscriptions(mail)
+      logger.debug "Sending to subscriptions."
+      self.subscriptions.each do |subscription|
+        begin
+          subscription.send_mail(mail)
+        rescue => exc
+          msg = I18n.t('errors.delivery_error',
+                       { email: subscription.email, error: exc.to_s })
+          logger.error msg
+        end
+      end
+    end
+
     private
 
     def set_gnupg_home
