@@ -1,4 +1,16 @@
 module Mail
+  # creates a Mail::Message likes schleuder
+  def self.create_message_to_list(msg, recipient, list)
+    mail = Mail.new(msg)
+    mail.list = list
+    mail.recipient = recipient
+    # don't freeze here, as the mail might not be fully
+    # parsed as body is lazy evaluated and might still
+    # be changed later.
+    mail.original_message = mail.dup #.freeze
+    mail
+  end
+
   # TODO: Test if subclassing breaks integration of mail-gpg.
   class Message
     attr_accessor :recipient
@@ -8,7 +20,7 @@ module Mail
     # TODO: This should be in initialize(), but I couldn't understand the
     # strange errors about wrong number of arguments when overriding
     # Message#initialize.
-    def setup(recipient, list)
+    def setup
       if self.encrypted?
         new = self.decrypt(verify: true)
         ## Work around a bug in mail-gpg: when decrypting pgp/mime the
@@ -25,10 +37,18 @@ module Mail
         new = self
       end
 
-      new.list = list
+      new.list = self.list
+      new.recipient = self.recipient
+
       new.gpg list.gpg_sign_options
       new.original_message = self.dup.freeze
-      new.recipient = recipient
+      # Trigger method early to save the information. Later some information
+      # might be gone (e.g. request-keywords that delete subscriptions or
+      # keys).
+      new.signer
+      self.dynamic_pseudoheaders.each do |str|
+        new.add_pseudoheader(str)
+      end
       new
     end
 
@@ -49,18 +69,28 @@ module Mail
         clean.add_part new_part
       end
 
-      # Attach body or mime-parts, respectively.
+      # Attach body or mime-parts in a new wrapper-part, to preserve the
+      # original mime-structure.
+      # We can't use self.to_s here — that includes all the headers we *don't*
+      # want to copy.
+      wrapper_part = Mail::Part.new
+      # Copy headers to are relevant for the mime-structure.
+      wrapper_part.content_type = self.content_type
+      wrapper_part.content_transfer_encoding = self.content_transfer_encoding if self.content_transfer_encoding
+      wrapper_part.content_disposition = self.content_disposition if self.content_disposition
+      wrapper_part.content_description = self.content_description if self.content_description
+      # Copy contents.
       if self.multipart?
         self.parts.each do |part|
-          clean.add_part Mail::Part.new(part)
+          wrapper_part.add_part(part)
         end
       else
-        # Don't use self.to_s here — that includes all the headers we *don't*
-        # want to copy.
-        new_part = Mail::Part.new
-        new_part.body = self.body.to_s
-        clean.add_part Mail::Part.new(new_part)
+        # We copied the content-headers, so we need to copy the body encoded.
+        # Otherwise the content might become unlegible.
+        wrapper_part.body = self.body.encoded
       end
+      clean.add_part(wrapper_part)
+
       clean
     end
 
@@ -106,8 +136,10 @@ module Mail
     end
 
     def signer
-      if signing_key.present?
-        list.subscriptions.where(fingerprint: signing_key.fingerprint).first
+      @signer ||= begin
+        if signing_key.present?
+          list.subscriptions.where(fingerprint: signing_key.fingerprint).first
+        end
       end
     end
 
@@ -122,10 +154,27 @@ module Mail
 
     def reply_to_signer(output)
       reply = self.reply
-      reply.body = Array(output).join("\n")
+      self.class.all_to_message_part(output).each do |part|
+        reply.add_part(part)
+      end
       self.signer.send_mail(reply)
     end
 
+    def self.all_to_message_part(input)
+      Array(input).map do |thing|
+        case thing
+        when Mail::Part
+          thing
+        when String
+          Mail::Part.new do
+            body thing.to_s
+          end
+        else
+          raise "Don't know how to handle input: #{thing.inspect}"
+        end
+      end
+    end
+
     def sendkey_request?
       @recipient.match(/-sendkey@/)
     end
@@ -138,12 +187,15 @@ module Mail
       @recipient.match(/-request@/)
     end
 
-    def bounce?
-      @recipient.match(/-bounce@/) ||
+    def automated_message?
+      @recipient.match(/-bounce@/).present? ||
           # Empty Return-Path
           self.return_path.to_s == '<>' ||
-          # Auto-Submitted exists and does not equal 'no'
-          ( self['Auto-Submitted'].present? && self['Auto-Submitted'].to_s.downcase != 'no' )
+          # Auto-Submitted exists and does not equal 'no' and no cron header
+          # present, as cron emails have the auto-submitted header.
+          ( self['Auto-Submitted'].present? && \
+            self['Auto-Submitted'].to_s.downcase != 'no' && \
+            !self['X-Cron-Env'].present?)
     end
 
     def keywords
@@ -162,7 +214,7 @@ module Mail
         end
         # TODO: Find multiline arguments (add-key). Currently add-key has to
         # read the whole body and hope for the best.
-        if line.match(/^x-([^: ]*)[: ]*(.*)/i)
+        if line.match(/^x-([^:\s]*)[:\s]*(.*)/i)
           command = $1.strip.downcase
           arguments = $2.to_s.strip.downcase.split(/[,; ]{1,}/)
           @keywords << [command, arguments]
@@ -187,9 +239,13 @@ module Mail
       _add_subject_prefix(:out)
     end
 
-    def add_pseudoheader(key, value)
+    def add_pseudoheader(string_or_key, value=nil)
       @dynamic_pseudoheaders ||= []
-      @dynamic_pseudoheaders << make_pseudoheader(key, value)
+      if value.present?
+        @dynamic_pseudoheaders << make_pseudoheader(string_or_key, value)
+      else
+        @dynamic_pseudoheaders << string_or_key.to_s
+      end
     end
 
     def make_pseudoheader(key, value)
@@ -305,7 +361,7 @@ module Mail
           pref << ')'
         end
 
-        fingerprint = list.key.fingerprint
+        fingerprint = list.fingerprint
         comment = "(Send an email to #{list.sendkey_address} to receive the public-key)"
 
         self['OpenPGP'] = "id=0x#{fingerprint} #{comment}; #{pref}"
@@ -326,6 +382,29 @@ module Mail
       end
     end
 
+    def first_plaintext_part(part=nil)
+      part ||= self
+      if part.multipart?
+        first_plaintext_part(part.parts.first)
+      elsif part.mime_type == 'text/plain'
+        part
+      else
+        nil
+      end
+    end
+
+
+    def attach_list_key!(list)
+      filename = "#{list.email}.asc"
+      self.add_file({
+        filename: filename,
+        content: list.export_key
+      })
+      self.attachments[filename].content_type = 'application/pgp-keys'
+      self.attachments[filename].content_description = 'OpenPGP public key'
+      true
+    end
+
     private
 
 
@@ -342,8 +421,10 @@ module Mail
       if ! string.empty?
         prefix = "#{string} "
         # Only insert prefix if it's not present already.
-        if ! self.subject.include?(prefix)
-          self.subject = "#{string} #{self.subject}"
+        if self.subject.nil?
+          self.subject = string
+        elsif ! self.subject.include?(prefix)
+          self.subject = "#{prefix}#{self.subject}"
         end
       end
     end
@@ -365,17 +446,6 @@ module Mail
       end
     end
 
-    def first_plaintext_part(part=nil)
-      part ||= self
-      if part.multipart?
-        first_plaintext_part(part.parts.first)
-      elsif part.mime_type == 'text/plain'
-        part
-      else
-        nil
-      end
-    end
-
     def clutch_anglebrackets(input)
       Array(input).map do |string|
         if string.first == '<'
